#!/usr/bin/php
<?php
/*
 *  Copyright (c) 2004-2005 Klaraelvdalens Datakonsult AB
 *
 *    Writen by Steffen Hansen <steffen@klaralvdalens-datakonsult.se>
 *
 *  This  program is free  software; you can redistribute  it and/or
 *  modify it  under the terms of the GNU  General Public License as
 *  published by the  Free Software Foundation; either version 2, or
 *  (at your option) any later version.
 *
 *  This program is  distributed in the hope that it will be useful,
 *  but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 *  General Public License for more details.
 *
 *  You can view the  GNU General Public License, online, at the GNU
 *  Project's homepage; see <http://www.gnu.org/licenses/gpl.html>.
 */

/* Fix include_path to pick up our modified Horde classes */
$include_path = ini_get('include_path');
ini_set( 'include_path', 
	 '.:/var/www/html/kolab:/usr/share/pear:'.$include_path);

require_once 'PEAR.php';
require_once '/var/www/html/kolab/kolabfilter/misc.php';
require_once '/var/www/html/kolab/kolabfilter/kolabmailtransport.php';

// Profiling code
/*
function mymtime(){
  $tmp=explode(' ',microtime());
  $rt=$tmp[0]+$tmp[1];
  return $rt;
}

class MyTimer {
  function MyTimer( $name ) {
    $this->name = $name;
  }
  function start() {
    $this->time = mymtime();
  }
  function stop() {
    $time = 100*(mymtime()-$this->time);
    myLog("Section ".$this->name." took $time msecs", RM_LOG_DEBUG);
  }

  var $name;
  var $time;
};

$totaltime =& new MyTimer("Total");
$totaltime->start();
*/

// Load our configuration file
$params = array();
require_once '/etc/kolab/resmgr/resmgr.conf';
init();

define( 'TMPDIR', '/var/lib/kolab/resmgr/filter' );
define( 'EX_TEMPFAIL', 75 );
define( 'EX_UNAVAILABLE', 69 );

//$inputtime =& new MyTimer("Input");
//$inputtime->start();

// Temp file for storing the message
$tmpfname = tempnam( TMPDIR, 'IN.' );
$tmpf = @fopen($tmpfname, "w");
if( !$tmpf ) {
  myLog("Error: Could not open $tempfname for writing: ".php_error(), RM_LOG_ERROR);
  exit(EX_TEMPFAIL);
}

// Cleanup function
function cleanup() {
  global $tmpfname;
  file_exists($tmpfname) && unlink($tmpfname);
}
register_shutdown_function( 'cleanup' );

function is_my_domain( $addr ) {
  global $params;
  if( is_array($params['email_domain']) ) {
	$domains = $params['email_domain'];
  } else {
	$domains = array($params['email_domain']);
  }
  
  $adrs = imap_rfc822_parse_adrlist($addr, $params['email_domain']);
  foreach ($adrs as $adr) {
    $adrdom = $adr->host;
    if( empty($adrdom) ) continue;
    foreach( $domains as $dom ) {
      if( $dom == $adrdom ) return true;
      if( $params['verify_subdomains'] && substr($adrdom, -strlen($dom)-1) == ".$dom" ) return true;
    }
  }
  return false;
}

/**
 Returns a list of allowed email addresses for user $sasluser
 or a PEAR_Error object if something croaked.
*/
function addrs_for_uid( $sasluser )
{
  global $params;
  /* Connect to the LDAP server and retrieve the users'
   allowed email addresses */
  $ldap = ldap_connect($params['ldap_uri']);
  if (!ldap_bind($ldap, $params['bind_dn'], $params['bind_pw'])) {
    myLog('Unable to contact LDAP server: ' . ldap_error($ldap));
    return new PEAR_Error('Unable to contact LDAP server: ' . ldap_error($ldap));
  }
  
  $filter = "(&(objectClass=kolabInetOrgPerson)(|(mail=$sasluser)(uid=$sasluser)))";
  $result = ldap_search($ldap, $params['base_dn'],
			$filter,
			array("dn", "mail", "alias" ));
  if (!$result) {
    myLog('Unable to perform LDAP search: ' . ldap_error($ldap));
    return new PEAR_Error('Unable to perform LDAP search: ' . ldap_error($ldap));
  }
  
  $entries = ldap_get_entries($ldap, $result);
  if ($entries['count'] != 1) {
    myLog($entries['count']." objects returned for uid $sasluser");
    return new PEAR_Error("Temporary LDAP error, unable to look up user $sasluser");
  }
  unset($entries[0]['mail']['count']);
  unset($entries[0]['alias']['count']);
  $addrs = array_merge((array) $entries[0]['mail'],(array) $entries[0]['alias']);
  $mail = $entries[0]['mail'][0];

  ldap_free_result($result);

  $filter = "(&(objectClass=kolabInetOrgPerson)(kolabDelegate=$mail))";
  $result = ldap_search($ldap, $params['base_dn'],
			$filter,
			array("dn", "mail" ));
  if (!$result) {
    myLog('Unable to perform LDAP search: ' . ldap_error($ldap));
    return new PEAR_Error('Unable to perform LDAP search: ' . ldap_error($ldap));
  }
  
  $entries = ldap_get_entries($ldap, $result);
  unset( $entries['count'] );
  foreach( $entries as $adr ) {
    if( $adr['mail']['count'] > 0 ) {
      unset($adr['mail']['count']);
      $addrs = array_merge((array) $addrs,(array) $adr['mail']);
    }
  }
  ldap_free_result($result);
  ldap_close($ldap);

  #myLog("Found addresses ".print_r($addrs,true)." for user $sasluser", RM_LOG_DEBUG);
  return $addrs;
}

/** Returns the format string used to rewrite
    the From header for untrusted messages */
function get_untrusted_subject_insert($sasluser,$sender)
{
  global $params;
  if( $sasluser ) {
    if( array_key_exists('untrusted_subject_insert', $params) ) {
      $fmt = $params['untrusted_subject_insert'];
    } else {
      $fmt = "(UNTRUSTED, sender is <%s>)";
    }
  } else {
    if( array_key_exists('unauthenticated_subject_insert', $params) ) {
      $fmt = $params['unauthenticated_subject_insert'];
    } else {
      $fmt = "(UNTRUSTED, sender <%s> is not authenticated)";
    }
  }
  return sprintf($fmt,$sender);
}

/** Check that the From header is not trying
    to impersonate a valid user that is not
    $sasluser. Returns one of:

    * True if From can be accepted
    * False if From must be rejected
    * A string with a corrected From header that makes
      From acceptable
    * A PEAR_Error object if something croaked
*/
function verify_sender( $sasluser, $sender, $fromhdr, $client_addr ) {
  global $params;

  /* Allow anything from localhost and
     fellow Kolab-hosts */
  if( $client_addr == '127.0.0.1' ) return true;
  $kolabhosts = split(',', $params['kolabhosts'] );
  $kolabhosts = array_map( "gethostbyname", $kolabhosts );
  if( array_search( $client_addr, $kolabhosts ) !== false ) return true;

  if( is_array($params['email_domain']) ) {
    $domains = $params['email_domain'];
  } else {
    $domains = array($params['email_domain']);
  }

  if( $sasluser ) {
    if( PEAR::isError($allowed_addrs = addrs_for_uid($sasluser)) ) {
      myLog("Error reading allowed addresses for $sasluser: ".$allowed_addrs->getMessage(), RM_LOG_ERROR);
      return $allowed_addrs;
    }
  } else {
    $allowed_addrs = false;
  }
  $untrusted = get_untrusted_subject_insert($sasluser,$sender);
  $adrs = imap_rfc822_parse_adrlist($fromhdr, $params['email_domain'][0]);
  foreach ($adrs as $adr) {
    $from = $adr->mailbox.'@'.$adr->host;
    $fromdom = $adr->host;
    if( $sasluser ) {
      if( !in_array( strtolower($from), $allowed_addrs ) ) {
	myLog("$from is not an allowed From address for $sasluser", RM_LOG_DEBUG);
	return false;
      }
    } else {
      foreach( $domains as $domain ) {
	if( strtolower($fromdom) == $domain 
	    || ( $params['verify_subdomains'] 
		 && substr($fromdom, -strlen($domain)-1) == ".$domain" ) ) {
	  if( $params['reject_forged_from_header'] ) {
	    myLog("$from is not an allowed From address for unauthenticated users", RM_LOG_DEBUG);	    
	    return false;
	  } else {
	    /* Rewrite */
	    myLog("$from is not an allowed From address for unauthenticated users, rewriting", RM_LOG_DEBUG);
	    
	    if( strpos( $fromhdr, $untrusted )===false ) {
	      return '"'.str_replace(array("\\",'"'),array("\\\\",'\"'),$adr->personal).' '.$untrusted.'" '.'<'.$from.'>';
	    } else {
	      return true;
	    }
	  }
	}
      }
    }
  }

  /* All seems OK */
  return true;


  /* TODO: What do we do about subdomains? */
  /*
    $senderdom = substr(strrchr($sender, '@'), 1);
    foreach( $domains as $domain ) {
      if( $params['verify_subdomains'] ) {	
	if( ($senderdom == $domain ||
	     $fromdom   == $domain ||
	     substr($senderdom, -strlen($domain)-1) == ".$domain" ||
	     substr($fromdom, -strlen($domain)-1) == ".$domain" ) &&
	    $sender != $from ) {
	  return false;
	}
      } else {
	if( ($senderdom == $domain ||
	     $fromdom   == $domain ) &&
	    $sender != $from ) {
		  return false;
	}
      }
    }
  }
  return true;
  */
}

$options = parse_args( array( 's', 'r', 'c', 'h', 'u' ), $_SERVER['argv']); //getopt("s:r:c:h:u:");

if (!array_key_exists('r', $options) || !array_key_exists('s', $options)) {
    fwrite(STDOUT, "Usage is $argv[0] -s sender@domain -r recip@domain\n");
    exit(EX_TEMPFAIL);
}

$sender = strtolower($options['s']);
$recipients = $options['r'];
$client_address = $options['c'];
$fqhostname = strtolower($options['h']);
$sasl_username = strtolower($options['u']);

// make sure recipients is an array
if( !is_array($recipients) ) {
  $recipients = array( $recipients );
}

// make recipients lowercase
for( $i = 0; $i < count($recipients); $i++ ) {
  $recipients[$i] = strtolower($recipients[$i]);
}

myLog("Kolabfilter starting up, user=$sasl_username, sender=$sender, recipients=".join(',', $recipients)
      .", client_address=$client_address", RM_LOG_DEBUG);

define( RM_STATE_READING_HEADER, 1 );
define( RM_STATE_READING_FROM,   2 );
define( RM_STATE_READING_SUBJECT,3 );
define( RM_STATE_READING_SENDER, 4 );
define( RM_STATE_READING_BODY,   5 );

$ical = false;
$from = false;
$subject = false;
$senderok = true;
$rewrittenfrom = false;
$state = RM_STATE_READING_HEADER;
while (!feof(STDIN) && $state != RM_STATE_READING_BODY) {
  $buffer = fgets(STDIN, 8192);
  $line = rtrim( $buffer, "\r\n");
  if( $line == '' ) {
    // Done with headers
    $state = RM_STATE_READING_BODY;
    if( $from && $params['verify_from_header'] ) {
      $rc = verify_sender( $sasl_username, $sender, $from, $client_address);
      if( PEAR::isError($rc) ) {
	$msg = $error->getMessage().", code ".$error->getCode();
	mylog($msg, RM_LOG_ERROR);
	fwrite(STDOUT, $msg."\n"); 
	exit(EX_TEMPFAIL);
      } else if( $rc === true ) {
	/* All OK, do nothing */
      } else if( $rc === false ) {
	/* Reject! */
	$senderok = false;
      } else if( is_string($rc) ) {
	/* Rewrite from */
	if( strpos( $from, $rc )===false ) {
	  myLog("Rewriting '$from' to '$rc'", RM_LOG_DEBUG);
	  $rewrittenfrom = "From: $rc\r\n";
	}
      }
    }
  } else {
    if( $line[0] != ' ' && $line[0] != "\t" ) $state = RM_STATE_READING_HEADER;
    switch( $state ) {
    case RM_STATE_READING_HEADER:
      if( $params['allow_sender_header'] && eregi( '^Sender: (.*)', $line, $regs ) ) {
	$from = $regs[1];
	$state = RM_STATE_READING_SENDER;
      } else if( !$from && eregi( '^From: (.*)', $line, $regs ) ) {
	$from = $regs[1];
	$state = RM_STATE_READING_FROM;
      } else if( eregi( '^Subject: (.*)', $line, $regs ) ) {
	$subject = $regs[1];
	$state = RM_STATE_READING_SUBJECT;
      } else if( eregi( '^Content-Type: text/calendar', $line ) ) {
	myLog("Found iCal data in message", RM_LOG_DEBUG);
	$ical = true;
      }
      break;
    case RM_STATE_READING_FROM:
      $from .= $line;
      break;
    case RM_STATE_READING_SENDER:
      $from .= $line;
      break;
    case RM_STATE_READING_SUBJECT:
      $subject .= $line;
      break;
    }
  }
  if( fwrite($tmpf, $buffer) === false ) {
    exit(EX_TEMPFAIL);
  }
}
while (!feof(STDIN)) {
  $buffer = fread( STDIN, 8192 );
  if( fwrite($tmpf, $buffer) === false ) {
    exit(EX_TEMPFAIL);
  }
}
fclose($tmpf);

//$inputtime->stop();

if( !$senderok ) {
  if( $ical && $params['allow_outlook_ical_forward'] ) {
    require_once('/var/www/html/kolab/kolabfilter/olhacks.php');
    $rc = olhacks_embedical( $fqhostname, $sender, $recipients, $from, $subject, $tmpfname );
    if( PEAR::isError( $rc ) ) {
      fwrite(STDOUT,"Filter failed: ".$rc->getMessage()."\n");
      exit(EX_TEMPFAIL);
    } else if( $rc === true ) {
      exit(0);
    }
  } else {
    $msg = "Invalid From: header. $from looks like a forged sender";
    myLog($msg, RM_LOG_DEBUG);
    fwrite(STDOUT,"$msg\n");
    exit(EX_UNAVAILABLE);
  }
}

//$outputtime =& new MyTimer("Output");
//$outputtime->start();

$tmpf = @fopen($tmpfname,"r");
if( !$tmpf ) {
  myLog("Error: Could not open $tempfname for reading: ".php_error(), RM_LOG_ERROR);
  exit(EX_TEMPFAIL);
}

/* TODO: Dont hardcode localhost:10025 */
$smtp = new KolabSMTP( 'localhost', 10025 );
if( PEAR::isError( $smtp ) ) {
  fwrite(STDOUT, $error->getMessage().", code ".$error->getCode()."\n"); 
  if( $error->getCode() < 500 ) exit(EX_TEMPFAIL);
  else exit(EX_UNAVAILABLE);
}
if( PEAR::isError( $error = $smtp->start($sender,$recipients) ) ) {
  fwrite(STDOUT, $error->getMessage().", code ".$error->getCode()."\n"); 
  if( $error->getCode() < 500 ) exit(EX_TEMPFAIL);
  else exit(EX_UNAVAILABLE);
}

$state = RM_STATE_READING_HEADER;
while (!feof($tmpf) && $state != RM_STATE_READING_BODY) {
  $buffer = fgets($tmpf, 8192);
  if( $rewrittenfrom ) {
    if( eregi( '^From: (.*)', $buffer ) ) {
      if( PEAR::isError($error = $smtp->data( $rewrittenfrom )) ) {
	$str = $error->getMessage().", code ".$error->getCode();
	myLog($str,RM_LOG_ERROR);
	fwrite(STDOUT, $str."\n");
	if( $error->getCode() < 500 ) exit(EX_TEMPFAIL);
	else exit(EX_UNAVAILABLE);
      }
      $state = RM_STATE_READING_FROM;
      continue;
    } else if( $state == RM_STATE_READING_FROM &&
	       ($buffer[0] == ' ' || $buffer[0] == "\t" ) ) {
      // Folded From header, ignore
      continue;
    }
  }
  if( rtrim( $buffer, "\r\n" ) == '' ) {
    $state = RM_STATE_READING_BODY;
  } else if( $buffer[0] != ' ' && $buffer[0] != "\t" ) {
    $state = RM_STATE_READING_HEADER;
  }
  if( PEAR::isError($error = $smtp->data( $buffer )) ) {
    $str = $error->getMessage().", code ".$error->getCode();
    myLog($str,RM_LOG_ERROR);
    fwrite(STDOUT, $str."\n");
    if( $error->getCode() < 500 ) exit(EX_TEMPFAIL);
    else exit(EX_UNAVAILABLE);
  }
}
while (!feof($tmpf) ) {
    $buffer = fread($tmpf, 8192);
    $len = strlen($buffer);

    /* We can't tolerate that the buffer breaks the data
       between \r and \n, so we try to avoid that. The limit 
       of 100 reads is to battle abuse */
    while( $buffer{$len-1} == "\r" && $len < 8192 + 100 ) {
      $buffer .= fread($tmpf,1);
      $len++;
    }
    if( PEAR::isError($error = $smtp->data( $buffer )) ) {
        fwrite(STDOUT, $error->getMessage().", code ".$error->getCode()."\n"); 
	if( $error->getCode() < 500 ) exit(EX_TEMPFAIL);
	else exit(EX_UNAVAILABLE);
    }
};

//myLog("Calling smtp->end()", RM_LOG_DEBUG);
$smtp->end();
//$outputtime->stop();
myLog("Kolabfilter successfully completed", RM_LOG_DEBUG);
//$totaltime->stop();
exit(0);
?>
